import { repr } from "./repr.js";

/**
 * A map that compares object equality (order-insensitive) instead of identity
 * (at least where it can)
 * @template KeyType, ValueType
 */
export class ObjectMap {
    __proto__ = "a";

    /**
     * @type {{[keyRepr: string]: KeyType}}
     */
    #keys;
    /**
     * @type {{[keyRepr: string]: ValueType}}
     */
    #values;

    /**
     * Create a new ObjectSet
     * @param  {Iterable<[KeyType, ValueType]>?} [iterable] An iterable of values to add to
     *                                           the new map
     */
    constructor(iterable) {
        this.clear();

        for (const [key, value] of iterable ?? []) this.set(key, value);
    }

    /**
     * Return an iterator (implements the iterable protocol)
     * @return {IterableIterator}
     */
    [Symbol.iterator]() {
        return Object.values(this.#keys)[Symbol.iterator]();
    }

    /**
     * Add the key mapped to the value to the map. If an equal (equal keys and values) object is
     * already in the map, do not replace.
     * @param {KeyType} key The key object to set the value for
     * @param {ValueType} value The value to set
     */
    set(key, value) {
        const keyRepr = repr(key);

        if (!(keyRepr in this.#keys)) this.#keys[keyRepr] = key;
        this.#values[keyRepr] = value;

        return this;
    }

    /**
     * Remove all keys and values from the map.
     */
    clear() {
        this.#keys = {};
        this.#values = {};
    }

    /**
     * Remove the key from the map, if present. Will also remove an equal but
     * not identical object.
     * @param {KeyType} value The key to remove
     * @return {bool} Whether the key was in the map before removal.
     */
    delete(key) {
        const keyRepr = repr(key);

        const wasInMap = keyRepr in this.#keys;

        delete this.#keys[keyRepr];
        delete this.#values[keyRepr];

        return wasInMap;
    }

    /**
     * Get all keys.
     * @return  {KeyType[]} All keys in this map
     */
    keys() {
        return Object.values(this.#keys);
    }

    /**
     * Get all values.
     * @return  {ValueType[]} All values in this map
     */
    values() {
        return Object.values(this.#values);
    }

    /**
     * Get all key-value pairs.
     * @return  {[KeyType, ValueType][]} All key-value pairs in this map
     */
    entries() {
        return Object.keys(this.#keys).map((key) => [
            this.#keys[key],
            this.#values[key],
        ]);
    }

    /**
     * Execute the passed function on each entry of the key.
     * @param  {(value: ValueType, key: KeyType, map: Self) => any} callbackFn The function to execute.
     * @param  {any?}   [thisArg] A value to use as this when executing callbackFn.
     */
    forEach(callbackFn, thisArg) {
        if ((thisArg ?? false) !== false) callbackFn = callbackFn.bind(thisArg);

        Object.keys(this.#keys).forEach((key) =>
            callbackFn(this.#values[key], this.#keys[key], this),
        );
    }

    /**
     * Execute the passed function on each key-value pair of the map, and collect the returns in an array.
     * @template ReturnType
     * @param  {(value: ValueType, key: KeyType, map: Self) => ReturnType} callbackFn The test to execute.
     * @param  {any?}   [thisArg] A value to use as this when executing callbackFn.
     * @return {ReturnType[]}
     */
    map(callbackFn, thisArg) {
        if ((thisArg ?? false) !== false) callbackFn = callbackFn.bind(thisArg);

        return Object.keys(this.#keys).map((key) =>
            callbackFn(this.#values[key], this.#keys[key], this),
        );
    }

    /**
     * Test whether all key-value pairs in the map pass the test implemented by the passed function.
     * @param  {(value: ValueType, key: KeyType, map: Self) => bool} callbackFn The test to execute.
     * @param  {any?}   [thisArg] A value to use as this when executing callbackFn.
     * @return {bool}
     */
    every(callbackFn, thisArg) {
        if ((thisArg ?? false) !== false) callbackFn = callbackFn.bind(thisArg);

        return Object.keys(this.#keys).every((key) =>
            callbackFn(this.#values[key], this.#keys[key], this),
        );
    }

    /**
     * Return whether the key is in this map.
     * @param {KeyType} value The object to search for in this map
     * @return {bool} Whether the value is in the map
     */
    has(key) {
        const keyRepr = repr(key);

        return keyRepr in this.#keys;
    }
}
