import { repr } from "./repr.js";

/**
 * A set that compares object equality (order-insensitive) instead of identity
 * (at least where it can)
 * @template ValueType
 */
export class ObjectSet {
    __proto__ = "a";

    /**
     * @type {{[valueRepr: string]: ValueType}}
     */
    #values;

    /**
     * Create a new ObjectSet
     * @param  {Iterable<ValueType>?} [iterable] An iterable of values to add to
     *                                           the new set
     */
    constructor(iterable) {
        this.clear();

        for (const value of iterable ?? []) this.add(value);
    }

    /**
     * Return an iterator (implements the iterable protocol)
     * @return {IterableIterator}
     */
    [Symbol.iterator]() {
        return Object.values(this.#values)[Symbol.iterator]();
    }

    /**
     * Alias for {@linkcode ObjectSet.values}
     * @return  {KeyType[]} All values in this set
     */
    keys() {
        return this.values();
    }

    /**
     * Get all values.
     * @return  {ValueType[]} All values in this set
     */
    values() {
        return Object.values(this.#values);
    }

    /**
     * Add the object to the set. If an equal (equal-repr) object is
     * already in the set, do not replace.
     *
     * @see {@linkcode repr}
     *
     * @param {ValueType} value The object to add
     */
    add(value) {
        const valueRepr = repr(value);

        if (!(valueRepr in this.#values)) this.#values[valueRepr] = value;

        return this;
    }

    /**
     * Remove all values from the set.
     */
    clear() {
        this.#values = {};
    }

    /**
     * Remove the object from the set, if present. Will also remove an equal but
     * not identical object. Uses repr for equality.
     *
     * @see {@linkcode repr}
     *
     * @param {ValueType} value The object to remove
     * @return {bool} Whether the value was in the set before removal.
     */
    delete(value) {
        const valueRepr = repr(value);

        const wasInSet = valueRepr in this.#values;

        delete this.#values[valueRepr];

        return wasInSet;
    }

    /**
     * Execute the passed function on each element of the set.
     * @param  {(value: ValueType, key: ValueType, set: Self) => any} callbackFn The function to execute.
     * @param  {any?}   [thisArg] A value to use as this when executing callbackFn.
     */
    forEach(callbackFn, thisArg) {
        if ((thisArg ?? false) !== false) callbackFn = callbackFn.bind(thisArg);

        Object.values(this.#values).forEach((value) =>
            callbackFn(value, value, this),
        );
    }

    /**
     * Execute the passed function on each element of the set, and collect the returns in an array.
     * @template ReturnType
     * @param  {(value: ValueType, key: ValueType, set: Self) => ReturnType} callbackFn The test to execute.
     * @param  {any?}   [thisArg] A value to use as this when executing callbackFn.
     * @return {ReturnType[]}
     */
    map(callbackFn, thisArg) {
        if ((thisArg ?? false) !== false) callbackFn = callbackFn.bind(thisArg);

        return Object.values(this.#values).map((value) =>
            callbackFn(value, value, this),
        );
    }

    /**
     * Test whether all elements in the set pass the test implemented by the passed function.
     * @param  {(value: ValueType, key: ValueType, set: Self) => bool} callbackFn The test to execute.
     * @param  {any?}   [thisArg] A value to use as this when executing callbackFn.
     * @return {bool}
     */
    every(callbackFn, thisArg) {
        if ((thisArg ?? false) !== false) callbackFn = callbackFn.bind(thisArg);

        return Object.values(this.#values).every((value) =>
            callbackFn(value, value, this),
        );
    }

    /**
     * Returns a set that only contains elements of this set that pass the test implemented by the passed function.
     * @param  {(value: ValueType, key: ValueType, set: Self) => bool} callbackFn The test to execute.
     * @param  {any?}   [thisArg] A value to use as this when executing callbackFn.
     * @return {ObjectSet<ValueType>}
     */
    filter(callbackFn, thisArg) {
        if ((thisArg ?? false) !== false) callbackFn = callbackFn.bind(thisArg);

        const filteredSet = new ObjectSet();

        this.forEach((value) =>
            callbackFn(value, value, this) ? filteredSet.add(value) : null,
        );

        return filteredSet;
    }

    /**
     * Return whether the value is an element of the set.
     * @param {ValueType} value The object to search for in this set
     * @return {bool} Whether the value is in the set
     */
    has(value) {
        const valueRepr = repr(value);

        return valueRepr in this.#values;
    }
}
