import { ObjectMap } from "./objectMap.js";
import { ObjectSet } from "./objectSet.js";

/**
 * A map that stores references of un"repr"able objects mapped to integer IDs.
 * @type {WeakMap<any, number>}
 */
const idMap = new WeakMap();
let currentId = 0;

const getIdFor = (value) => {
    if (idMap.has(value)) return idMap.get(value);

    const id = currentId++;
    idMap.set(value, id);
    return id;
};

/**
 * Create a canonical string representation for any* value.
 *
 * This is especially handy to define a somewhat rigid object equality.
 * Assumes absolutely no mutation to passed in objects, ever. Oh look,
 * it says "gullible" on the ceiling.
 *
 * @param  {any} value The value to represent
 * @return {string} The value's string representation
 */
export const repr = (value) => {
    if (value === Object.prototype) {
        return "Object.prototype";
    }

    if (typeof value === "function") {
        return `function ${value.name}[${getIdFor(value)}](${value.length} ${value.length === 1 ? "arg" : "args"
            }) { /* ... */ }`;
    }

    if (typeof value !== "object") {
        if (Number.isNaN(value)) return "NaN";
        if (value === undefined) return "undefined";

        return (
            (JSON.stringify(value) ?? String(value)) ||
            "undefined /* unhashable */"
        );
    }

    // if value is a global prototype, use its common name
    if (
        "constructor" in value &&
        typeof value.constructor === "function" &&
        value.constructor.name in globalThis &&
        value === globalThis[value.constructor.name].prototype
    ) {
        return `${value.constructor.name}.prototype`;
    }

    if (value instanceof Set || value instanceof ObjectSet) {
        return `{__proto__: ${repr(
            Object.getPrototypeOf(value),
        )}; __elements__: [${[...value.values()]
            .map(repr)
            .toSorted()
            .join(", ")}]${Object.getOwnPropertyNames(value).length > 0 ? "; " : ""
            }${Object.getOwnPropertyNames(value)
                .toSorted()
                .filter((n) => {
                    try {
                        value[n];
                        return true;
                    } catch {
                        return false;
                    }
                })
                .map((name) => `${name}: ${repr(value[name])}`)
                .join(", ")}}`;
    }

    if (value instanceof Map || value instanceof ObjectMap) {
        return `{__proto__: ${repr(
            Object.getPrototypeOf(value),
        )}; __elements__: {${[...value.keys()]
            .toSorted()
            .map((k) => [k, value.get(k)])
            .map(([k, v]) => `${k}: ${repr(v)}`)
            .join(", ")}}${Object.getOwnPropertyNames(value).length > 0 ? "; " : ""
            }${Object.getOwnPropertyNames(value)
                .toSorted()
                .filter((n) => {
                    try {
                        value[n];
                        return true;
                    } catch {
                        return false;
                    }
                })
                .map((name) => `${name}: ${repr(value[name])}`)
                .join(", ")}}`;
    }

    if (
        value instanceof WeakRef ||
        value instanceof WeakMap ||
        value instanceof WeakMap
    ) {
        return `{__proto__: ${repr(
            Object.getPrototypeOf(value),
        )}; __id__: {${getIdFor(value)}}`;
    }

    return `{__proto__: ${repr(Object.getPrototypeOf(value))}${Object.getOwnPropertyNames(value).length > 0 ? "; " : ""
        }${Object.getOwnPropertyNames(value)
            .toSorted()
            .filter((n) => {
                try {
                    value[n];
                    return true;
                } catch {
                    return false;
                }
            })
            .map((name) => `${name}: ${repr(value[name])}`)
            .join(", ")}}`;
};
